import UIKit
import MapKit

class AddUserServiceController: UIViewController , MKMapViewDelegate, CLLocationManagerDelegate,  UIPickerViewDelegate, UIPickerViewDataSource {

    private var serviceTypes = ["Reforms", "Transport", "Home", "New Technologies", "Care", "Professors"]
    private var center : CLLocationCoordinate2D?
    
    @IBOutlet weak var newServiceTitle: UITextField!
    @IBOutlet weak var newServiceDescription: UITextField!
    @IBOutlet weak var newServicePrice: UITextField!
    @IBOutlet weak var newServiceLocation: MKMapView!
    @IBOutlet weak var serviceTypePiker: UIPickerView!
    
    var locationManager : CLLocationManager?
    
    override func viewDidLoad() {
        super.viewDidLoad()
        
        let addServiceButton = UIBarButtonItem(title: "Done", style: .plain, target: self, action: #selector(addService))
        navigationItem.rightBarButtonItems = [addServiceButton]
        
        self.configureMap()
        self.configureLocationManager()
        self.configureServiceTypePiker()

        self.title = "Add new service"
        
        let tapGesture = UITapGestureRecognizer(target: self, action: #selector(self.dismissKeyboard))
        self.view.addGestureRecognizer(tapGesture)
        
        //This notificated and executes selectors when the keyboard appears/disappear.
        NotificationCenter.default.addObserver(self, selector: #selector(self.showedKeyboard), name:NSNotification.Name.UIKeyboardWillShow, object: nil);
        NotificationCenter.default.addObserver(self, selector: #selector(self.hidedKeyboard), name:NSNotification.Name.UIKeyboardWillHide, object: nil);
    }
    
    /**
    * Function which generate a request to server to add a new service.
    */
    @objc func addService() {
        
        let selectedServiceType = self.serviceTypes[self.serviceTypePiker.selectedRow(inComponent: 0)]
        let selectedServiceTypeId = getSelectedServiceTypeByName(selectedServiceType)
        
        var longitude : Double?
        var latitude : Double?
        
        if (self.center == nil) {
            longitude =  -3.020793
            latitude = 39.837116
        } else {
            longitude = (self.center?.longitude)!
            latitude = (self.center?.latitude)!
        }
        
        let createServiceRequest = CreateServiceRequest(
            ServiceType: selectedServiceTypeId,
            ServiceName: self.newServiceTitle.text!,
            ServiceDescription: self.newServiceDescription.text!,
            Price: Int(self.newServicePrice.text!)!,
            Longitude: longitude!,
            Latitude: latitude!)
        
        SerlocServiceConsumer.createNewService(Request: createServiceRequest, Controller: self)
    }
    
    /**
    * Configure the service type piker.
    */
    private func configureServiceTypePiker() {
        self.serviceTypePiker.delegate = self
        self.serviceTypePiker.dataSource = self
    }
    
    /**
    * Function which configures the view map.
    */
    private func configureMap() {
        self.newServiceLocation.isZoomEnabled = true
        self.newServiceLocation.isScrollEnabled = true
        
        self.newServiceLocation.delegate = self
        self.newServiceLocation.showsUserLocation = false
    }
    
    /**
    * Return the number of components of the date picker.
    */
    func numberOfComponents(in pickerView: UIPickerView) -> Int {
        return 1
    }
    
    /**
    * Return the number of elements displayer on the service type piker.
    */
    func pickerView(_ pickerView: UIPickerView, numberOfRowsInComponent component: Int) -> Int {
        return self.serviceTypes.count
    }
    
    /**
    * Return the values displayed on the service type piker.
    */
    func pickerView(_ pickerView: UIPickerView, titleForRow row: Int, forComponent component: Int) -> String? {
        return self.serviceTypes[row]
    }
    
    /**
     * Mehtod to tell to the map view that user location has changed.
     */
    func mapView(_  mapView: MKMapView, didUpdate userLocation: MKUserLocation) {
        let mapSpan = MKCoordinateSpan(latitudeDelta: 0.05, longitudeDelta: 0.05)
        let actualUserRegion = MKCoordinateRegion(center: userLocation.coordinate , span: mapSpan)
        self.newServiceLocation?.setRegion(actualUserRegion, animated: true)
    }
    
    /**
    * Function which traks all the map movements to update the actual selected position.
    */
    func mapView(_ mapView: MKMapView, regionDidChangeAnimated animated: Bool) {
        self.center = mapView.centerCoordinate
    }
    
    /**
     * Function to configure the location manager.
     */
    private func configureLocationManager() {
        self.locationManager = CLLocationManager()
        self.locationManager?.delegate = self
        self.locationManager?.allowsBackgroundLocationUpdates = true
        self.locationManager?.distanceFilter = 100
        self.locationManager?.desiredAccuracy = kCLLocationAccuracyBest
        
        let status : CLAuthorizationStatus = CLLocationManager.authorizationStatus()
        if status == CLAuthorizationStatus.notDetermined || status == CLAuthorizationStatus.denied || status == CLAuthorizationStatus.authorizedWhenInUse {
            
            self.locationManager?.requestAlwaysAuthorization()
            self.locationManager?.requestWhenInUseAuthorization()
        }
        self.locationManager?.startUpdatingLocation()
        self.locationManager?.startUpdatingHeading()
    }

    override func didReceiveMemoryWarning() {
        super.didReceiveMemoryWarning()
        // Dispose of any resources that can be recreated.
    }
    
    /**
     * Function to return to the user services list.
     */
    public func uploadAndRedirectToUserServicesList() {
        _ = navigationController?.popViewController(animated: true)
    }

    /**
    * Get the selected service type id by the service type name.
    */
    private func getSelectedServiceTypeByName(_ typeName : String) -> Int {
        
        switch typeName {
            case "Reforms":
                return 1
            case "Transport":
                return 2
            case "Home":
                return 3
            case "New Technologies":
                return 4
            case "Care":
                return 5
            case "Professors":
                return 6
            default:
                return 0
        }
    }
    
    /**
     *Function which hides keyboard on tap on screen.
     */
    @objc public func dismissKeyboard (_ sender: UITapGestureRecognizer) {
        newServiceTitle.resignFirstResponder()
        newServiceDescription.resignFirstResponder()
        newServicePrice.resignFirstResponder()
    }
    
    @objc func showedKeyboard(sender: NSNotification) {
        if (newServicePrice.isEditing) {
         self.view.frame.origin.y = -150
        }
    }
    
    @objc func hidedKeyboard(sender: NSNotification) {
        self.view.frame.origin.y = 0
    }
}
