import UIKit
import MapKit

class ServicesMapController: UIViewController, MKMapViewDelegate, CLLocationManagerDelegate {

    @IBOutlet weak var mapView: MKMapView!
    public var services : Array<Service> = []
    public var servicesArray : [Service] = [Service]()
    public var selectedService : Service?
    public var previousSelectedServiceType : Int?
    
    var locationManager : CLLocationManager?
    
    override func viewDidLoad() {
        super.viewDidLoad()
        
        servicesArray.reserveCapacity(services.count)
        services.forEach { (service) in
            servicesArray.append(service)
        }

        self.configureMapView()
        self.configureLocationManager()
        self.addServicesMarks(Services: servicesArray)
        
        // Do any additional setup after loading the view.
    }
    
    /**
     * Recovering tab bar if returning from service detail view.
     */
    override func viewWillAppear(_ animated: Bool) {
        self.tabBarController?.tabBar.isHidden = false
    }

    override func didReceiveMemoryWarning() {
        super.didReceiveMemoryWarning()
        // Dispose of any resources that can be recreated.
    }
    
    /**
     * Function to configure the map view.
     */
    private func configureMapView() {
        self.mapView.isZoomEnabled = true
        self.mapView.isScrollEnabled = true
        
        self.mapView.delegate = self
        self.mapView.showsUserLocation = true
    }
    
    /**
     * Function to configure the location manager.
     */
    private func configureLocationManager() {
        self.locationManager = CLLocationManager()
        self.locationManager?.delegate = self
        self.locationManager?.allowsBackgroundLocationUpdates = true
        self.locationManager?.distanceFilter = 100
        self.locationManager?.desiredAccuracy = kCLLocationAccuracyBest
        
        let status : CLAuthorizationStatus = CLLocationManager.authorizationStatus()
        if status == CLAuthorizationStatus.notDetermined || status == CLAuthorizationStatus.denied || status == CLAuthorizationStatus.authorizedWhenInUse {
            
            self.locationManager?.requestAlwaysAuthorization()
            self.locationManager?.requestWhenInUseAuthorization()
        }
        self.locationManager?.startUpdatingLocation()
        self.locationManager?.startUpdatingHeading()
    }
    
    /**
     * Function which add marks for all the service definitions.
     */
    private func addServicesMarks(Services services : Array<Service>) {
        for service in services {
        
            let serviceLongitude = CLLocationDegrees(service.longitude)
            let serviceLatitude = CLLocationDegrees(service.latitude)
            
            let serviceCoordinate : CLLocationCoordinate2D = CLLocationCoordinate2D(latitude: serviceLatitude, longitude: serviceLongitude)
            
            let serviceMark = ServiceMapMark(id: service.id, coordinate: serviceCoordinate, title: service.name, subtitle: service.description, price: service.price)
            self.mapView.addAnnotation(serviceMark)
        }
    }
    
    /**
     * Mehtod to tell to the map view that user location has changed.
     */
    func mapView(_  mapView: MKMapView, didUpdate userLocation: MKUserLocation) {
        let mapSpan = MKCoordinateSpan(latitudeDelta: 0.05, longitudeDelta: 0.05)
        let actualUserRegion = MKCoordinateRegion(center: userLocation.coordinate , span: mapSpan)
        self.mapView?.setRegion(actualUserRegion, animated: true)
    }
    
    /**
    * Function which determines the behaviout when some marker is selected
    */
    func mapView(_ mapView: MKMapView, didSelect view: MKAnnotationView) {
        let selectedServiceMark = view.annotation as! ServiceMapMark
        
        servicesArray.forEach { (service) in
            if (service.id == selectedServiceMark.id) {
                self.selectedService = service
            }
        }
        
        performSegue(withIdentifier: "showServiceDetail", sender: self)
    }
    
    
    /**
    * When the map view is dragged or moved we need to refresh the markers to display
    * all the available services with the new visible map area.
    */
    func mapView(_ mapView: MKMapView, regionDidChangeAnimated animated: Bool) {
        let nortToSouthDistance = mapView.region.span.latitudeDelta
        let estToWestDistance = mapView.region.span.longitudeDelta
        
        let actualPosition = mapView.region.center
        
        let newMaxLatitude : Float = Float(actualPosition.latitude + (nortToSouthDistance.binade / 2))
        let newMinLatitude : Float = Float(actualPosition.latitude - (nortToSouthDistance.binade / 2))
        let newMaxLongitude : Float = Float(actualPosition.longitude + (estToWestDistance.binade / 2))
        let newminLongitude : Float = Float(actualPosition.longitude - (estToWestDistance.binade / 2))
        
        let updateServicesRequest : UpdateServicesRequest = UpdateServicesRequest(MaxLatitude: newMaxLatitude, MinLatitude: newMinLatitude, MaxLongitude: newMaxLongitude, MinLongitude: newminLongitude, ServiceType: self.previousSelectedServiceType!)
        
        SerlocServiceConsumer.updateServicesMapVisibleServices(Request: updateServicesRequest, Map: self.mapView, Controller: self)
        
    }
    
    override func prepare(for segue: UIStoryboardSegue, sender: Any?) {
        if (segue.identifier == "showServiceDetail") {
            (segue.destination as! ServiceDetailController).serviceDetail = selectedService!
        }
    }
}
